@@ -12,9 +12,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library. If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -30,12 +28,19 @@
 #include "gtkmodulesprivate.h"
 #include "gtkborderimageprivate.h"
 #include "gtkpango.h"
-#include "gtkshadowprivate.h"
+#include "gtkcssarrayvalueprivate.h"
+#include "gtkcsscornervalueprivate.h"
+#include "gtkcssenumvalueprivate.h"
+#include "gtkcssnumbervalueprivate.h"
+#include "gtkcssrgbavalueprivate.h"
+#include "gtkcssshadowsvalueprivate.h"
 #include "gtkcsstypesprivate.h"
 #include "gtkthemingengineprivate.h"
 #include "gtkroundedboxprivate.h"
 #include "gtkthemingbackgroundprivate.h"
 
+#include "fallback-c89.c"
+
 /**
  * SECTION:gtkthemingengine
  * @Short_description: Theming renderers
@@ -203,6 +208,8 @@ struct GtkThemingModuleClass
 #define GTK_THEMING_MODULE(o)    (G_TYPE_CHECK_INSTANCE_CAST ((o), GTK_TYPE_THEMING_MODULE, GtkThemingModule))
 #define GTK_IS_THEMING_MODULE(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GTK_TYPE_THEMING_MODULE))
 
+GType gtk_theming_module_get_type (void);
+
 G_DEFINE_TYPE (GtkThemingModule, gtk_theming_module, G_TYPE_TYPE_MODULE);
 
 static void
@@ -330,14 +337,31 @@ _gtk_theming_engine_set_context (GtkThemingEngine *engine,
   priv->context = context;
 }
 
-const GValue *
-_gtk_theming_engine_peek_property (GtkThemingEngine *engine,
-                                   const char       *property_name)
+GtkStyleContext *
+_gtk_theming_engine_get_context (GtkThemingEngine *engine)
 {
   g_return_val_if_fail (GTK_IS_THEMING_ENGINE (engine), NULL);
-  g_return_val_if_fail (property_name != NULL, NULL);
 
-  return _gtk_style_context_peek_property (engine->priv->context, property_name);
+  return engine->priv->context;
+}
+
+GtkCssValue *
+_gtk_theming_engine_peek_property (GtkThemingEngine *engine,
+                                   guint             property_id)
+{
+  g_return_val_if_fail (GTK_IS_THEMING_ENGINE (engine), NULL);
+
+  return _gtk_style_context_peek_property (engine->priv->context, property_id);
+}
+
+double
+_gtk_theming_engine_get_number (GtkThemingEngine *engine,
+                                guint             property_id,
+                                double            one_hundred_percent)
+{
+  g_return_val_if_fail (GTK_IS_THEMING_ENGINE (engine), 0.0);
+
+  return _gtk_style_context_get_number (engine->priv->context, property_id, one_hundred_percent);
 }
 
 /**
@@ -558,18 +582,17 @@ gtk_theming_engine_get_state (GtkThemingEngine *engine)
  * Returns: %TRUE if there is a running transition animation for @state.
  *
  * Since: 3.0
+ *
+ * Deprecated: 3.6: Always returns %FALSE
  **/
 gboolean
 gtk_theming_engine_state_is_running (GtkThemingEngine *engine,
                                      GtkStateType      state,
                                      gdouble          *progress)
 {
-  GtkThemingEnginePrivate *priv;
-
   g_return_val_if_fail (GTK_IS_THEMING_ENGINE (engine), FALSE);
 
-  priv = engine->priv;
-  return gtk_style_context_state_is_running (priv->context, state, progress);
+  return FALSE;
 }
 
 /**
@@ -760,7 +783,7 @@ gtk_theming_engine_get_border_color (GtkThemingEngine *engine,
 
 /**
  * gtk_theming_engine_get_border:
- * @engine: a #GtkthemingEngine
+ * @engine: a #GtkThemingEngine
  * @state: state to retrieve the border for
  * @border: (out): return value for the border settings
  *
@@ -783,7 +806,7 @@ gtk_theming_engine_get_border (GtkThemingEngine *engine,
 
 /**
  * gtk_theming_engine_get_padding:
- * @engine: a #GtkthemingEngine
+ * @engine: a #GtkThemingEngine
  * @state: state to retrieve the padding for
  * @padding: (out): return value for the padding settings
  *
@@ -1029,10 +1052,8 @@ gtk_theming_engine_render_check (GtkThemingEngine *engine,
   gtk_theming_engine_get_color (engine, flags, &fg_color);
   gtk_theming_engine_get_background_color (engine, flags, &bg_color);
   gtk_theming_engine_get_border (engine, flags, &border);
-
-  gtk_theming_engine_get (engine, flags,
-                          "border-style", &border_style,
-                          NULL);
+  border_style = _gtk_css_border_style_value_get 
+    (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_TOP_STYLE));
 
   border_width = MIN (MIN (border.top, border.bottom),
                       MIN (border.left, border.right));
@@ -1085,22 +1106,14 @@ gtk_theming_engine_render_check (GtkThemingEngine *engine,
     }
   else
     {
-      gdouble progress;
-      gboolean running;
-
-      running = gtk_theming_engine_state_is_running (engine, GTK_STATE_ACTIVE, &progress);
-
-      if ((flags & GTK_STATE_FLAG_ACTIVE) || running)
+      if (flags & GTK_STATE_FLAG_ACTIVE)
         {
-          if (!running)
-            progress = 1;
-
           cairo_translate (cr,
                            x + pad, y + pad);
 
           cairo_scale (cr, interior_size / 7., interior_size / 7.);
 
-          cairo_rectangle (cr, 0, 0, 7 * progress, 7);
+          cairo_rectangle (cr, 0, 0, 7, 7);
           cairo_clip (cr);
 
           cairo_move_to  (cr, 7.0, 0.0);
@@ -1162,10 +1175,8 @@ gtk_theming_engine_render_option (GtkThemingEngine *engine,
   gtk_theming_engine_get_color (engine, flags, &fg_color);
   gtk_theming_engine_get_background_color (engine, flags, &bg_color);
   gtk_theming_engine_get_border (engine, flags, &border);
-
-  gtk_theming_engine_get (engine, flags,
-                          "border-style", &border_style,
-                          NULL);
+  border_style = _gtk_css_border_style_value_get 
+    (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_TOP_STYLE));
 
   exterior_size = MIN (width, height);
   border_width = MIN (MIN (border.top, border.bottom),
@@ -1663,6 +1674,7 @@ render_border (cairo_t       *cr,
                                      border->left - other_border.left);
             render_frame_fill (cr, &other_box, &other_border, colors, dont_draw);
           }
+          break;
         case GTK_BORDER_STYLE_GROOVE:
         case GTK_BORDER_STYLE_RIDGE:
           {
@@ -1727,12 +1739,8 @@ render_frame_internal (GtkThemingEngine *engine,
   GtkStateFlags state;
   GtkBorderStyle border_style[4];
   GtkRoundedBox border_box;
-  gdouble progress;
-  gboolean running;
   GtkBorder border;
-  GdkRGBA *alloc_colors[4];
   GdkRGBA colors[4];
-  guint i;
 
   state = gtk_theming_engine_get_state (engine);
 
@@ -1743,75 +1751,33 @@ render_frame_internal (GtkThemingEngine *engine,
     _gtk_border_image_render (&border_image, &border, cr, x, y, width, height);
   else
     {
-      gtk_theming_engine_get (engine, state,
-                              "border-top-style", &border_style[0],
-                              "border-right-style", &border_style[1],
-                              "border-bottom-style", &border_style[2],
-                              "border-left-style", &border_style[3],
-                              "border-top-color", &alloc_colors[0],
-                              "border-right-color", &alloc_colors[1],
-                              "border-bottom-color", &alloc_colors[2],
-                              "border-left-color", &alloc_colors[3],
-                              NULL);
+      border_style[0] = _gtk_css_border_style_value_get (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_TOP_STYLE));
+      border_style[1] = _gtk_css_border_style_value_get (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_RIGHT_STYLE));
+      border_style[2] = _gtk_css_border_style_value_get (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_BOTTOM_STYLE));
+      border_style[3] = _gtk_css_border_style_value_get (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_LEFT_STYLE));
 
-      running = gtk_theming_engine_state_is_running (engine, GTK_STATE_PRELIGHT, &progress);
-
-      if (running)
-        {
-          GtkStateFlags other_state;
-          GdkRGBA *other_colors[4];
-
-          if (state & GTK_STATE_FLAG_PRELIGHT)
-            {
-              other_state = state & ~(GTK_STATE_FLAG_PRELIGHT);
-              progress = 1 - progress;
-            }
-          else
-            other_state = state | GTK_STATE_FLAG_PRELIGHT;
-
-          gtk_theming_engine_get (engine, other_state,
-                                  "border-top-color", &other_colors[0],
-                                  "border-right-color", &other_colors[1],
-                                  "border-bottom-color", &other_colors[2],
-                                  "border-left-color", &other_colors[3],
-                                  NULL);
-
-          for (i = 0; i < 4; i++)
-            {
-              colors[i].red = CLAMP (alloc_colors[i]->red + ((other_colors[i]->red - alloc_colors[i]->red) * progress), 0, 1);
-              colors[i].green = CLAMP (alloc_colors[i]->green + ((other_colors[i]->green - alloc_colors[i]->green) * progress), 0, 1);
-              colors[i].blue = CLAMP (alloc_colors[i]->blue + ((other_colors[i]->blue - alloc_colors[i]->blue) * progress), 0, 1);
-              colors[i].alpha = CLAMP (alloc_colors[i]->alpha + ((other_colors[i]->alpha - alloc_colors[i]->alpha) * progress), 0, 1);
-              gdk_rgba_free (other_colors[i]);
-              gdk_rgba_free (alloc_colors[i]);
-            }
-        }
-      else
-        {
-          for (i = 0; i < 4; i++)
-            {
-              colors[i] = *alloc_colors[i];
-              gdk_rgba_free (alloc_colors[i]);
-            }
-        }
+      colors[0] = *_gtk_css_rgba_value_get_rgba (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_TOP_COLOR));
+      colors[1] = *_gtk_css_rgba_value_get_rgba (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_RIGHT_COLOR));
+      colors[2] = *_gtk_css_rgba_value_get_rgba (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_BOTTOM_COLOR));
+      colors[3] = *_gtk_css_rgba_value_get_rgba (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_LEFT_COLOR));
 
       _gtk_rounded_box_init_rect (&border_box, x, y, width, height);
-      _gtk_rounded_box_apply_border_radius (&border_box, engine, state, junction);
+      _gtk_rounded_box_apply_border_radius_for_engine (&border_box, engine, junction);
 
       render_border (cr, &border_box, &border, hidden_side, colors, border_style);
     }
 
-  border_style[0] = g_value_get_enum (_gtk_theming_engine_peek_property (engine, "outline-style"));
+  border_style[0] = _gtk_css_border_style_value_get (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_OUTLINE_STYLE));
   if (border_style[0] != GTK_BORDER_STYLE_NONE)
     {
       int offset;
 
       border_style[1] = border_style[2] = border_style[3] = border_style[0];
-      border.top = g_value_get_int (_gtk_theming_engine_peek_property (engine, "outline-width"));
+      border.top = round (_gtk_css_number_value_get (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_OUTLINE_WIDTH), 100));
       border.left = border.right = border.bottom = border.top;
-      colors[0] = *(GdkRGBA *) g_value_get_boxed (_gtk_theming_engine_peek_property (engine, "outline-color"));
+      colors[0] = *_gtk_css_rgba_value_get_rgba (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_OUTLINE_COLOR));
       colors[3] = colors[2] = colors[1] = colors[0];
-      offset = g_value_get_int (_gtk_theming_engine_peek_property (engine, "outline-offset"));
+      offset = _gtk_css_number_value_get (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_OUTLINE_OFFSET), 100);
       
       /* reinit box here - outlines don't have a border radius */
       _gtk_rounded_box_init_rect (&border_box, x, y, width, height);
@@ -1861,7 +1827,7 @@ gtk_theming_engine_render_expander (GtkThemingEngine *engine,
   double x_double, y_double;
   gdouble angle;
   gint line_width;
-  gboolean running, is_rtl;
+  gboolean is_rtl;
   gdouble progress;
 
   cairo_save (cr);
@@ -1870,12 +1836,9 @@ gtk_theming_engine_render_expander (GtkThemingEngine *engine,
   gtk_theming_engine_get_color (engine, flags, &fg_color);
   gtk_theming_engine_get_border_color (engine, flags, &outline_color);
 
-  running = gtk_theming_engine_state_is_running (engine, GTK_STATE_ACTIVE, &progress);
   is_rtl = (gtk_theming_engine_get_direction (engine) == GTK_TEXT_DIR_RTL);
   line_width = 1;
-
-  if (!running)
-    progress = (flags & GTK_STATE_FLAG_ACTIVE) ? 1 : 0;
+  progress = (flags & GTK_STATE_FLAG_ACTIVE) ? 1 : 0;
 
   if (!gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_HORIZONTAL))
     {
@@ -2081,49 +2044,16 @@ gtk_theming_engine_render_layout (GtkThemingEngine *engine,
                                   PangoLayout      *layout)
 {
   GdkRGBA fg_color;
-  GtkShadow *text_shadow = NULL;
   GtkStateFlags flags;
-  gdouble progress;
-  gboolean running;
 
   cairo_save (cr);
   flags = gtk_theming_engine_get_state (engine);
   gtk_theming_engine_get_color (engine, flags, &fg_color);
 
-  running = gtk_theming_engine_state_is_running (engine, GTK_STATE_PRELIGHT, &progress);
-
-  if (running)
-    {
-      GtkStateFlags other_flags;
-      GdkRGBA other_fg;
-
-      if (flags & GTK_STATE_FLAG_PRELIGHT)
-        {
-          other_flags = flags & ~(GTK_STATE_FLAG_PRELIGHT);
-          progress = 1 - progress;
-        }
-      else
-        other_flags = flags | GTK_STATE_FLAG_PRELIGHT;
-
-      gtk_theming_engine_get_color (engine, other_flags, &other_fg);
-
-      fg_color.red = CLAMP (fg_color.red + ((other_fg.red - fg_color.red) * progress), 0, 1);
-      fg_color.green = CLAMP (fg_color.green + ((other_fg.green - fg_color.green) * progress), 0, 1);
-      fg_color.blue = CLAMP (fg_color.blue + ((other_fg.blue - fg_color.blue) * progress), 0, 1);
-      fg_color.alpha = CLAMP (fg_color.alpha + ((other_fg.alpha - fg_color.alpha) * progress), 0, 1);
-    }
-
-  gtk_theming_engine_get (engine, flags,
-                          "text-shadow", &text_shadow,
-                          NULL);
-
   prepare_context_for_layout (cr, x, y, layout);
 
-  if (text_shadow != NULL)
-    {
-      _gtk_text_shadow_paint_layout (text_shadow, cr, layout);
-      _gtk_shadow_unref (text_shadow);
-    }
+  _gtk_css_shadows_value_paint_layout (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_TEXT_SHADOW),
+                                       cr, layout);
 
   gdk_cairo_set_source_rgba (cr, &fg_color);
   pango_cairo_show_layout (cr, layout);
@@ -2140,32 +2070,8 @@ gtk_theming_engine_render_slider (GtkThemingEngine *engine,
                                   gdouble           height,
                                   GtkOrientation    orientation)
 {
-  const GtkWidgetPath *path;
-  gint thickness;
-
-  path = gtk_theming_engine_get_path (engine);
-
   gtk_theming_engine_render_background (engine, cr, x, y, width, height);
   gtk_theming_engine_render_frame (engine, cr, x, y, width, height);
-
-  /* FIXME: thickness */
-  thickness = 2;
-
-  if (gtk_widget_path_is_type (path, GTK_TYPE_SCALE))
-    {
-      if (orientation == GTK_ORIENTATION_VERTICAL)
-        gtk_theming_engine_render_line (engine, cr,
-                                        x + thickness,
-                                        y + (gint) height / 2,
-                                        x + width - thickness - 1,
-                                        y + (gint) height / 2);
-      else
-        gtk_theming_engine_render_line (engine, cr,
-                                        x + (gint) width / 2,
-                                        y + thickness,
-                                        x + (gint) width / 2,
-                                        y + height - thickness - 1);
-    }
 }
 
 static void
@@ -2182,8 +2088,7 @@ gtk_theming_engine_render_frame_gap (GtkThemingEngine *engine,
   GtkJunctionSides junction;
   GtkStateFlags state;
   gint border_width;
-  GtkCssBorderCornerRadius *top_left_radius, *top_right_radius;
-  GtkCssBorderCornerRadius *bottom_left_radius, *bottom_right_radius;
+  GtkCssValue *corner[4];
   gdouble x0, y0, x1, y1, xc, yc, wc, hc;
   GtkBorder border;
 
@@ -2192,12 +2097,10 @@ gtk_theming_engine_render_frame_gap (GtkThemingEngine *engine,
   junction = gtk_theming_engine_get_junction_sides (engine);
 
   gtk_theming_engine_get_border (engine, state, &border);
-  gtk_theming_engine_get (engine, state,
-			  "border-top-left-radius", &top_left_radius,
-			  "border-top-right-radius", &top_right_radius,
-			  "border-bottom-right-radius", &bottom_right_radius,
-			  "border-bottom-left-radius", &bottom_left_radius,
-			  NULL);
+  corner[GTK_CSS_TOP_LEFT] = _gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_TOP_LEFT_RADIUS);
+  corner[GTK_CSS_TOP_RIGHT] = _gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_TOP_RIGHT_RADIUS);
+  corner[GTK_CSS_BOTTOM_LEFT] = _gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_BOTTOM_LEFT_RADIUS);
+  corner[GTK_CSS_BOTTOM_RIGHT] = _gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_BORDER_BOTTOM_RIGHT_RADIUS);
 
   border_width = MIN (MIN (border.top, border.bottom),
                       MIN (border.left, border.right));
@@ -2212,10 +2115,10 @@ gtk_theming_engine_render_frame_gap (GtkThemingEngine *engine,
       wc = MAX (xy1_gap - xy0_gap - 2 * border_width, 0);
       hc = border_width;
 
-      if (xy0_gap < top_left_radius->horizontal)
+      if (xy0_gap < _gtk_css_corner_value_get_x (corner[GTK_CSS_TOP_LEFT], width))
         junction |= GTK_JUNCTION_CORNER_TOPLEFT;
 
-      if (xy1_gap > width - top_right_radius->horizontal)
+      if (xy1_gap > width - _gtk_css_corner_value_get_x (corner[GTK_CSS_TOP_RIGHT], width))
         junction |= GTK_JUNCTION_CORNER_TOPRIGHT;
       break;
     case GTK_POS_BOTTOM:
@@ -2224,10 +2127,10 @@ gtk_theming_engine_render_frame_gap (GtkThemingEngine *engine,
       wc = MAX (xy1_gap - xy0_gap - 2 * border_width, 0);
       hc = border_width;
 
-      if (xy0_gap < bottom_left_radius->horizontal)
+      if (xy0_gap < _gtk_css_corner_value_get_x (corner[GTK_CSS_BOTTOM_LEFT], width))
         junction |= GTK_JUNCTION_CORNER_BOTTOMLEFT;
 
-      if (xy1_gap > width - bottom_right_radius->horizontal)
+      if (xy1_gap > width - _gtk_css_corner_value_get_x (corner[GTK_CSS_BOTTOM_RIGHT], width))
         junction |= GTK_JUNCTION_CORNER_BOTTOMRIGHT;
 
       break;
@@ -2237,10 +2140,10 @@ gtk_theming_engine_render_frame_gap (GtkThemingEngine *engine,
       wc = border_width;
       hc = MAX (xy1_gap - xy0_gap - 2 * border_width, 0);
 
-      if (xy0_gap < top_left_radius->vertical)
+      if (xy0_gap < _gtk_css_corner_value_get_y (corner[GTK_CSS_TOP_LEFT], height))
         junction |= GTK_JUNCTION_CORNER_TOPLEFT;
 
-      if (xy1_gap > height - bottom_left_radius->vertical)
+      if (xy1_gap > height - _gtk_css_corner_value_get_y (corner[GTK_CSS_BOTTOM_LEFT], height))
         junction |= GTK_JUNCTION_CORNER_BOTTOMLEFT;
 
       break;
@@ -2250,10 +2153,10 @@ gtk_theming_engine_render_frame_gap (GtkThemingEngine *engine,
       wc = border_width;
       hc = MAX (xy1_gap - xy0_gap - 2 * border_width, 0);
 
-      if (xy0_gap < top_right_radius->vertical)
+      if (xy0_gap < _gtk_css_corner_value_get_y (corner[GTK_CSS_TOP_RIGHT], height))
         junction |= GTK_JUNCTION_CORNER_TOPRIGHT;
 
-      if (xy1_gap > height - bottom_right_radius->vertical)
+      if (xy1_gap > height - _gtk_css_corner_value_get_y (corner[GTK_CSS_BOTTOM_RIGHT], height))
         junction |= GTK_JUNCTION_CORNER_BOTTOMRIGHT;
 
       break;
@@ -2271,11 +2174,6 @@ gtk_theming_engine_render_frame_gap (GtkThemingEngine *engine,
                          0, junction);
 
   cairo_restore (cr);
-
-  g_free (top_left_radius);
-  g_free (top_right_radius);
-  g_free (bottom_right_radius);
-  g_free (bottom_left_radius);
 }
 
 static void
@@ -2291,62 +2189,35 @@ gtk_theming_engine_render_extension (GtkThemingEngine *engine,
   GtkJunctionSides junction = 0;
   guint hidden_side = 0;
 
-  cairo_save (cr);
-
   switch (gap_side)
     {
     case GTK_POS_LEFT:
       junction = GTK_JUNCTION_LEFT;
       hidden_side = (1 << GTK_CSS_LEFT);
-
-      cairo_translate (cr, x + width, y);
-      cairo_rotate (cr, G_PI / 2);
       break;
     case GTK_POS_RIGHT:
       junction = GTK_JUNCTION_RIGHT;
       hidden_side = (1 << GTK_CSS_RIGHT);
-
-      cairo_translate (cr, x, y + height);
-      cairo_rotate (cr, - G_PI / 2);
       break;
     case GTK_POS_TOP:
       junction = GTK_JUNCTION_TOP;
       hidden_side = (1 << GTK_CSS_TOP);
-
-      cairo_translate (cr, x + width, y + height);
-      cairo_rotate (cr, G_PI);
       break;
     case GTK_POS_BOTTOM:
       junction = GTK_JUNCTION_BOTTOM;
       hidden_side = (1 << GTK_CSS_BOTTOM);
-
-      cairo_translate (cr, x, y);
       break;
     }
 
-  if (gap_side == GTK_POS_TOP ||
-      gap_side == GTK_POS_BOTTOM)
-    _gtk_theming_background_init (&bg, engine, 
-                                  0, 0,
-                                  width, height,
-                                  GTK_JUNCTION_BOTTOM);
-  else
-    _gtk_theming_background_init (&bg, engine, 
-                                  0, 0,
-                                  height, width,
-                                  GTK_JUNCTION_BOTTOM);
-
+  _gtk_theming_background_init (&bg, engine, 
+                                x, y,
+                                width, height,
+                                junction);
   _gtk_theming_background_render (&bg, cr);
 
-  cairo_restore (cr);
-
-  cairo_save (cr);
-
   render_frame_internal (engine, cr,
                          x, y, width, height,
                          hidden_side, junction);
-
-  cairo_restore (cr);
 }
 
 static void
@@ -2393,6 +2264,7 @@ gtk_theming_engine_render_handle (GtkThemingEngine *engine,
   GtkJunctionSides sides;
   GtkThemingBackground bg;
   gint xx, yy;
+  gboolean has_image;
 
   cairo_save (cr);
   flags = gtk_theming_engine_get_state (engine);
@@ -2405,8 +2277,11 @@ gtk_theming_engine_render_handle (GtkThemingEngine *engine,
   color_shade (&bg_color, 1.3, &lighter);
 
   _gtk_theming_background_init (&bg, engine, x, y, width, height, sides);
+  has_image = _gtk_theming_background_has_background_image (&bg);
   _gtk_theming_background_render (&bg, cr);
 
+  gtk_theming_engine_render_frame (engine, cr, x, y, width, height);
+
   if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_GRIP))
     {
       /* reduce confusing values to a meaningful state */
@@ -2656,12 +2531,15 @@ gtk_theming_engine_render_handle (GtkThemingEngine *engine,
     }
   else if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_PANE_SEPARATOR))
     {
-      if (width > height)
-        for (xx = x + width / 2 - 15; xx <= x + width / 2 + 15; xx += 5)
-          render_dot (cr, &lighter, &darker, xx, y + height / 2 - 1, 3);
-      else
-        for (yy = y + height / 2 - 15; yy <= y + height / 2 + 15; yy += 5)
-          render_dot (cr, &lighter, &darker, x + width / 2 - 1, yy, 3);
+      if (!has_image)
+        {
+          if (width > height)
+            for (xx = x + width / 2 - 15; xx <= x + width / 2 + 15; xx += 5)
+              render_dot (cr, &lighter, &darker, xx, y + height / 2 - 1, 3);
+          else
+            for (yy = y + height / 2 - 15; yy <= y + height / 2 + 15; yy += 5)
+              render_dot (cr, &lighter, &darker, x + width / 2 - 1, yy, 3);
+        }
     }
   else
     {
@@ -2677,10 +2555,10 @@ gtk_theming_engine_render_handle (GtkThemingEngine *engine,
 }
 
 void
-_gtk_theming_engine_paint_spinner (cairo_t *cr,
-                                   gdouble  radius,
-                                   gdouble  progress,
-                                   GdkRGBA *color)
+_gtk_theming_engine_paint_spinner (cairo_t       *cr,
+                                   gdouble        radius,
+                                   gdouble        progress,
+                                   const GdkRGBA *color)
 {
   guint num_steps, step;
   gdouble half;
@@ -2737,37 +2615,25 @@ render_spinner (GtkThemingEngine *engine,
                 gdouble           height)
 {
   GtkStateFlags state;
-  GtkShadow *shadow;
   GdkRGBA color;
-  gdouble progress;
   gdouble radius;
 
   state = gtk_theming_engine_get_state (engine);
-
-  if (!gtk_theming_engine_state_is_running (engine, GTK_STATE_ACTIVE, &progress))
-    progress = -1;
-
   radius = MIN (width / 2, height / 2);
 
   gtk_theming_engine_get_color (engine, state, &color);
-  gtk_theming_engine_get (engine, state,
-                          "icon-shadow", &shadow,
-                          NULL);
 
   cairo_save (cr);
   cairo_translate (cr, x + width / 2, y + height / 2);
 
-  if (shadow != NULL)
-    {
-      _gtk_icon_shadow_paint_spinner (shadow, cr,
-                                      radius,
-                                      progress);
-      _gtk_shadow_unref (shadow);
-    }
+  _gtk_css_shadows_value_paint_spinner (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_ICON_SHADOW),
+                                        cr,
+                                        radius,
+                                        -1);
 
   _gtk_theming_engine_paint_spinner (cr,
                                      radius,
-                                     progress,
+                                     -1,
                                      &color);
 
   cairo_restore (cr);
@@ -2781,13 +2647,18 @@ gtk_theming_engine_render_activity (GtkThemingEngine *engine,
                                     gdouble           width,
                                     gdouble           height)
 {
-  if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_SPINNER))
+  GtkThemingBackground bg;
+  
+  _gtk_theming_background_init (&bg, engine, x, y, width, height, 0);
+  
+  if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_SPINNER) &&
+      !_gtk_theming_background_has_background_image (&bg))
     {
       render_spinner (engine, cr, x, y, width, height);
     }
   else
     {
-      gtk_theming_engine_render_background (engine, cr, x, y, width, height);
+      _gtk_theming_background_render (&bg, cr);
       gtk_theming_engine_render_frame (engine, cr, x, y, width, height);
     }
 }
@@ -2930,24 +2801,11 @@ gtk_theming_engine_render_icon (GtkThemingEngine *engine,
                                 gdouble x,
                                 gdouble y)
 {
-  GtkStateFlags state;
-  GtkShadow *icon_shadow;
-
-  state = gtk_theming_engine_get_state (engine);
-
   cairo_save (cr);
 
   gdk_cairo_set_source_pixbuf (cr, pixbuf, x, y);
 
-  gtk_theming_engine_get (engine, state,
-                          "icon-shadow", &icon_shadow,
-                          NULL);
-
-  if (icon_shadow != NULL)
-    {
-      _gtk_icon_shadow_paint (icon_shadow, cr);
-      _gtk_shadow_unref (icon_shadow);
-    }
+  _gtk_css_shadows_value_paint_icon (_gtk_theming_engine_peek_property (engine, GTK_CSS_PROPERTY_ICON_SHADOW), cr);
 
   cairo_paint (cr);
 
